from oscar.apps.catalogue import views as oscar
from oscar.core.loading import get_class

from ocyan.core.fender import config
from ocyan.core.utils import merge_dicts

AddToBasketForm = get_class("basket.forms", "AddToBasketForm")


class ProductDetailView(oscar.ProductDetailView):
    enforce_parent = config.get("oscar", "enforce_parent", False)

    def get_basket_form(self):
        return AddToBasketForm(basket=self.request.basket, product=self.object)

    def get_media(self):
        return self.get_basket_form().media + self.get_alert_form().media

    def get_context_data(self, **kwargs):
        ctx = super(ProductDetailView, self).get_context_data(**kwargs)
        return merge_dicts(ctx, {"media": self.get_media()})



# Generated by Django 2.2.9 on 2020-03-19 15:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0020_merge_20191003_0948'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='option',
            options={'ordering': ('order', 'id'), 'verbose_name': 'Option', 'verbose_name_plural': 'Options'},
        ),
        migrations.AddField(
            model_name='option',
            name='order',
            field=models.IntegerField(blank=True, null=True, verbose_name='Ordering', help_text="Controls the ordering of product options on product detail pages"),
        ),
    ]



from decimal import Decimal

from django.db import models
from django.db.models.functions import Coalesce

from ocyan.plugin.oscar_catalogue.catalogue import abstract_models as ocyan


class AbstractProduct(ocyan.AbstractProduct):
    @property
    def variants_ordered_by_price(self):
        return (
            self.children.filter(is_public=True)
            .annotate(
                lowest_price=Coalesce(
                    models.Avg("stockrecords__price"),
                    models.Value(
                        Decimal(999999999)
                    ),  # put children with no stock at the end
                )
            )
            .order_by("lowest_price", "-priority")
        )

    @property
    def variants_ordered_for_admin(self):
        return (
            self.children.all()
            .annotate(
                lowest_price=Coalesce(
                    models.Avg("stockrecords__price"),
                    models.Value(
                        Decimal(999999999)
                    ),  # put children with no stock at the end
                )
            )
            .order_by("lowest_price", "-priority")
        )

    class Meta(ocyan.AbstractProduct.Meta):
        abstract = True



# Generated by Django 2.2.4 on 2019-08-16 08:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0016_auto_20190327_0757'),
    ]

    operations = [
    ]



# Generated by Django 2.2.13 on 2021-07-27 11:12

from django.db import migrations

def forward(apps, schema_editor):
    Category = apps.get_model("catalogue", "Category")
    Product = apps.get_model("catalogue", "Product")

    for c in Category._base_manager.all():
        if not c.meta_title:
            c.meta_title = c.seo_title
        if not c.meta_description:
            c.meta_description = c.seo_description

        c.save()

    for p in Product._base_manager.all():
        if not p.meta_title:
            p.meta_title = p.seo_title
        if not p.meta_description:
            p.meta_description = p.seo_description

        p.save()


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0023_oscar3_merge'),
    ]

    operations = [
        migrations.RunPython(forward, migrations.RunPython.noop),
    ]



# Generated by Django 4.2.11 on 2025-04-24 07:37

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("catalogue", "0030_merge_product_code"),
        ("catalogue", "0031_productcategoryhierarchy"),
    ]

    operations = []



# Generated by Django 3.2.23 on 2024-01-09 15:12

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0027_attributeoption_code_attributeoptiongroup_code_and_more'),
        ('catalogue', '0027_merge_20220708_1223'),
    ]

    operations = [
    ]



# pylint: disable=W0614,W0401
from oscar.core.loading import get_classes, is_model_registered

(
    AbstractCategory,
    AbstractProduct,
    AbstractOption,
    AbstractProductClass,
    AbstractProductAttribute,
    AbstractProductAttributeValue,
) = get_classes(
    "catalogue.abstract_models",
    [
        "AbstractCategory",
        "AbstractProduct",
        "AbstractOption",
        "AbstractProductClass",
        "AbstractProductAttribute",
        "AbstractProductAttributeValue",
    ],
)

###############################################################################
# This is how to override an oscar app, while keeping the overriding app      #
# overridable itself.                                                         #
# The mechanism is the same as in oscar, the only think different is the way  #
# ``__all__`` is being updated with the exports of the original oscar app.    #
###############################################################################

__all__ = []

if not is_model_registered("catalogue", "ProductClass"):

    class ProductClass(AbstractProductClass):
        pass

    __all__.append("ProductClass")


if not is_model_registered("catalogue", "Category"):

    class Category(AbstractCategory):
        pass

    __all__.append("Category")


if not is_model_registered("catalogue", "Product"):

    class Product(AbstractProduct):
        pass

    __all__.append("Product")


if not is_model_registered("catalogue", "Option"):

    class Option(AbstractOption):
        pass

    __all__.append("Option")


if not is_model_registered("catalogue", "ProductAttribute"):

    class ProductAttribute(AbstractProductAttribute):
        pass

    __all__.append("ProductAttribute")


if not is_model_registered("catalogue", "ProductAttributeValue"):

    class ProductAttributeValue(AbstractProductAttributeValue):
        pass

    __all__.append("ProductAttributeValue")

# both these lines are reruired, the first one imports all objects the are not
# implemented above into the current namespace.
# the second line updates the __all__ variable of the current namespace with
# the objects from the oscar namespace.
# THIS ENSURE THAT THE SAME CAN BE DONE BY ANOTHER APP, SO THIS APP CAN BE ONCE
# MORE EXTENDED BY ANOTHER APP.

from oscar.apps.catalogue.models import *
from oscar.apps.catalogue.models import __all__ as some

__all__.extend(some)



# Generated by Django 4.2.11 on 2024-12-04 10:45

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("catalogue", "0029_merge_20240911_1251"),
        ("catalogue", "0029_product_code"),
    ]

    operations = []



from django.conf import settings
from django.urls import path, re_path
from django.utils.translation import gettext, override
from django.views.generic.base import RedirectView

from ocyan.plugin.oscar_catalogue.catalogue.shared.views import (
    CatalogueIndexRedirectView,
    CategoryRedirectView,
)
from ocyan.plugin.oscar_catalogue.constants import SLUG_ID_SEPARATOR


def get_categories_as_disjunction():
    unique_categories = set(["category"])
    if len(settings.LANGUAGES) < 20:
        for lang in dict(settings.LANGUAGES).keys():
            with override(lang):
                unique_categories.add(gettext("category/")[:-1])

    return "|".join(unique_categories)


class CatalogueAppConfigMixin:
    """
    Because we're supporting <2.0 and >2.0 of oscar,
    we have to deal with their django app config implementation introduced in 2.0.

    In this mixin we define the urls to avoid duplicate code.
    """

    def new_urls(self):
        return [
            path(
                "<slug:product_slug>%s<int:pk>/" % SLUG_ID_SEPARATOR,
                self.detail_view.as_view(),
                name="detail",
            ),
            path(gettext("category/"), self.catalogue_view.as_view(), name="index"),
            re_path(
                r"^category/(?P<category_slug>[\w-]+(/[\w-]+)*){SLUG_ID_SEPARATOR}(?P<pk>\d+)/$".format(
                    SLUG_ID_SEPARATOR=SLUG_ID_SEPARATOR,
                ),
                self.category_view.as_view(),
                name="category",
            ),
            re_path(
                r"^({CATEGORIES_AS_DISJUNCTION})/(?P<category_slug>[\w-]+(/[\w-]+)*){SLUG_ID_SEPARATOR}(?P<pk>\d+)/$".format(
                    CATEGORIES_AS_DISJUNCTION=get_categories_as_disjunction(),
                    SLUG_ID_SEPARATOR=SLUG_ID_SEPARATOR,
                ),
                CategoryRedirectView.as_view(),
                name="redirect_category",
            ),
            re_path(
                r"^({CATEGORIES_AS_DISJUNCTION})/$".format(
                    CATEGORIES_AS_DISJUNCTION=get_categories_as_disjunction()
                ),
                CatalogueIndexRedirectView.as_view(),
                name="redirect_browse",
            ),
            path("ranges/<slug:slug>/", self.range_view.as_view(), name="range"),
        ]

    def old_urls(self):
        if SLUG_ID_SEPARATOR == "_":
            return []
        else:
            return [
                path(
                    "<slug:product_slug>_<int:pk>/",
                    RedirectView.as_view(
                        pattern_name="catalogue:detail",
                        permanent=True,
                        query_string=True,
                    ),
                ),
                re_path(
                    r"^category/(?P<category_slug>[\w-]+(/[\w-]+)*)_(?P<pk>\d+)/$",
                    RedirectView.as_view(
                        pattern_name="catalogue:category",
                        permanent=True,
                        query_string=True,
                    ),
                ),
            ]



# Generated by Django 2.2.13 on 2021-07-27 11:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0024_use_oscar_seo'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='option',
            options={'ordering': ['order', 'name'], 'verbose_name': 'Option', 'verbose_name_plural': 'Options'},
        ),
        migrations.RemoveField(
            model_name='category',
            name='seo_description',
        ),
        migrations.RemoveField(
            model_name='category',
            name='seo_keywords',
        ),
        migrations.RemoveField(
            model_name='category',
            name='seo_title',
        ),
        migrations.RemoveField(
            model_name='product',
            name='seo_description',
        ),
        migrations.RemoveField(
            model_name='product',
            name='seo_keywords',
        ),
        migrations.RemoveField(
            model_name='product',
            name='seo_title',
        ),
        migrations.RenameField(
            model_name='category',
            old_name='parents_are_public',
            new_name='ancestors_are_public'
        ),
        migrations.AlterField(
            model_name='category',
            name='ancestors_are_public',
            field=models.BooleanField(db_index=True, default=True, help_text='The ancestors of this category are public', verbose_name='Ancestor categories are public'),
        ),
    ]



# Generated by Django 2.1.9 on 2019-10-03 09:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0017_category_is_public'),
    ]

    operations = [
        migrations.AddField(
            model_name='category',
            name='parents_are_public',
            field=models.BooleanField(default=True, help_text='The parents of this category are public', verbose_name='Parent categories are public'),
        ),
    ]



# Generated by Django 2.2.13 on 2021-07-20 12:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0021_auto_20201005_0844'),
        ('catalogue', '0022_order_index'),
    ]

    operations = [
    ]



from django.contrib import admin
from django.db import models
from django.db.models.functions import Coalesce
from django.utils.safestring import mark_safe

from oscar.core.loading import get_model

Category = get_model("catalogue", "Category")


class CategoryFilter(admin.SimpleListFilter):
    title = "Categories"
    parameter_name = "category"

    def lookups(self, request, model_admin):
        value = self.value()
        qs = Category.objects.get_queryset()

        depth = 1
        if value is not None:
            _category = qs.get(pk=value)
            depth = _category.depth + 1

        return [
            (cat.id, mark_safe(("&nbsp;" * cat.depth) + cat.name))
            for cat in qs.filter(depth__lte=depth)
        ]

    def queryset(self, request, queryset):
        value = self.value()
        if value is not None:
            _category = Category.objects.get(pk=value)
            categories = Category.get_tree(_category)
            return queryset.filter(categories__in=categories).distinct()
        return queryset


class ProductClassFilter(admin.SimpleListFilter):
    title = "Producttype"
    parameter_name = "product_class"

    def lookups(self, request, model_admin):
        qs = model_admin.get_queryset(request)
        return set(
            qs.annotate(
                real_product_class_id=Coalesce(
                    models.F("product_class"), models.F("parent__product_class")
                ),
                real_product_class_name=Coalesce(
                    models.F("product_class__name"),
                    models.F("parent__product_class__name"),
                ),
            ).values_list("real_product_class_id", "real_product_class_name")
        )

    def queryset(self, request, queryset):
        value = self.value()
        qs = queryset.annotate(
            real_product_class=Coalesce(
                models.F("product_class"), models.F("parent__product_class")
            )
        )
        if value is not None:
            return qs.filter(real_product_class=value)
        return queryset



from django.contrib import admin

from oscar.apps.catalogue import admin as oscar
from oscar.apps.catalogue.admin import *  # noqa pylint: disable=W0611,W0401,W0614

from .filters import CategoryFilter, ProductClassFilter


# pylint: disable=function-redefined
class AttributeInline(oscar.AttributeInline):
    autocomplete_fields = ["attribute"]


# pylint: disable=function-redefined
class CategoryInline(oscar.CategoryInline):
    autocomplete_fields = ("category",)


# pylint: disable=function-redefined
class ProductAdmin(oscar.ProductAdmin):
    inlines = [AttributeInline, CategoryInline, oscar.ProductRecommendationInline]
    list_filter = oscar.ProductAdmin.list_filter + [
        "is_public",
        CategoryFilter,
        ProductClassFilter,
    ]
    readonly_fields = oscar.ProductAdmin.readonly_fields + ("product_class",)


class OptionAdmin(admin.ModelAdmin):
    list_display = ["name", "code", "type"]
    list_filter = ["type", "productclass"]
    search_fields = [
        "name",
        "code",
        "productclass__name",
        "productclass__slug",
        "product__upc",
        "product__slug",
        "product__title",
    ]


class ProductAttributeAdmin(oscar.ProductAttributeAdmin):
    search_fields = (
        "name",
        "code",
        "type",
        "product_class__name",
        "option_group__name",
    )


class ProductAttributeValueAdmin(oscar.ProductAttributeValueAdmin):
    autocomplete_fields = ("product", "attribute")


class ProductImageAdmin(admin.ModelAdmin):
    autocomplete_fields = ("product",)
    search_fields = ("caption", "product__upc", "product__title", "product__slug")


class CategoryAdmin(oscar.CategoryAdmin):
    search_fields = ("name", "slug", "pk")


class ProductCategoryAdmin(admin.ModelAdmin):
    list_display = ("pk", "product", "category")
    search_fields = (
        "product__upc",
        "product__title",
        "product__slug",
        "category__name",
        "category__slug",
    )
    autocomplete_fields = ("product", "category")


admin.site.unregister(Product)
admin.site.register(Product, ProductAdmin)
admin.site.unregister(ProductAttribute)
admin.site.register(ProductAttribute, ProductAttributeAdmin)
admin.site.unregister(ProductAttributeValue)
admin.site.register(ProductAttributeValue, ProductAttributeValueAdmin)
admin.site.unregister(ProductImage)
admin.site.register(ProductImage, ProductImageAdmin)
admin.site.unregister(Category)
admin.site.register(Category, CategoryAdmin)
admin.site.unregister(ProductCategory)
admin.site.register(ProductCategory, ProductCategoryAdmin)
admin.site.unregister(Option)
admin.site.register(Option, OptionAdmin)



from ocyan.core.plugins import OcyanPluginConfig
from ocyan.core.utils import merge_dicts


class OcyanOscar3CatalogueAppPluginConfig(OcyanPluginConfig):
    def get_smoketest_modules(self):
        return ["ocyan.plugin.oscar_catalogue.tests"]

    def contribute_to_settings(self, settings):
        return merge_dicts(
            settings,
            {
                "MIGRATION_MODULES": {
                    "catalogue": "ocyan.plugin.oscar_catalogue.oxyan_catalogue_extension.migrations"
                }
            },
        )

    def contribute_to_installed_apps(self, installed_apps):
        installed_apps.oscar.append(
            "ocyan.plugin.oscar_catalogue.catalogue.apps.CatalogueConfig",
            replaces="oscar.apps.catalogue.apps.CatalogueConfig",
        )
        installed_apps.oscar.register_extension(
            "ocyan.plugin.oscar_catalogue.oxyan_catalogue_extension",
            "ocyan.plugin.oscar_catalogue.catalogue",
        )
        installed_apps.ocyan.append("ocyan.plugin.oscar_catalogue")
        return installed_apps

    def contribute_to_i18n_fields(self, fields_dict):
        fields_dict = super().contribute_to_i18n_fields(fields_dict)
        fields_dict.update(
            {
                "catalogue.product.long_description": True,
                "catalogue.product.meta_title": False,
                "catalogue.product.meta_description": False,
                "catalogue.category.meta_title": False,
                "catalogue.category.meta_description": False,
            }
        )
        return fields_dict



# Generated by Django 2.0.13 on 2019-07-11 12:27

import django.core.validators
from django.db import migrations, models
from ocyan.plugin.oscar.fender import ALL_TAX_RATES, get_default_tax_rate as vat_default

class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0017_category_is_public'),
    ]

    operations = [
        migrations.AddField(
            model_name='category',
            name='seo_description',
            field=models.CharField(blank=True, help_text='SEO Meta descriptions should be less than 300 characters, if left empty the product description will be used', max_length=1024, null=True, verbose_name='SEO Meta description'),
        ),
        migrations.AddField(
            model_name='category',
            name='seo_keywords',
            field=models.CharField(blank=True, help_text="Meta Keywords don't make any improvements in SEO, use Meta description instead.", max_length=1024, null=True, verbose_name='SEO Meta keywords'),
        ),
        migrations.AddField(
            model_name='category',
            name='seo_title',
            field=models.CharField(blank=True, help_text='SEO Meta titles should be less than 60 characters, if left empty the product title will be used', max_length=255, null=True, verbose_name='SEO Meta title'),
        ),
        migrations.AddField(
            model_name='product',
            name='long_description',
            field=models.TextField(blank=True, help_text='A far more elaborate, but optional description of the product', null=True, verbose_name='Content'),
        ),
        migrations.AddField(
            model_name='product',
            name='seo_description',
            field=models.CharField(blank=True, help_text='SEO Meta descriptions should be less than 300 characters, if left empty the product description will be used', max_length=1024, null=True, verbose_name='SEO Meta description'),
        ),
        migrations.AddField(
            model_name='product',
            name='seo_keywords',
            field=models.CharField(blank=True, help_text="Meta Keywords don't make any improvements in SEO, use Meta description instead.", max_length=1024, null=True, verbose_name='SEO Meta keywords'),
        ),
        migrations.AddField(
            model_name='product',
            name='seo_title',
            field=models.CharField(blank=True, help_text='SEO Meta titles should be less than 60 characters, if left empty the product title will be used', max_length=255, null=True, verbose_name='SEO Meta title'),
        ),
        migrations.AddField(
            model_name='product',
            name='vat_rate',
            field=models.CharField(choices=ALL_TAX_RATES, default=vat_default(), max_length=5, null=True, verbose_name='VAT rate'),
        ),
        migrations.AddField(
            model_name='productattributevalue',
            name='value_embed',
            field=models.TextField(blank=True, help_text='This can be found in the URL of a video. So for example: https://www.youtube.com/watch?v=dkIU-ygvtps. In this case {dkIU-ygvtps} is the Youtube Video ID', null=True, verbose_name='Youtube Video ID'),
        ),
        migrations.AlterField(
            model_name='productattribute',
            name='type',
            field=models.CharField(choices=[('text', 'Text'), ('integer', 'Integer'), ('boolean', 'True / False'), ('float', 'Float'), ('richtext', 'Rich Text'), ('date', 'Date'), ('datetime', 'Datetime'), ('option', 'Option'), ('multi_option', 'Multi Option'), ('entity', 'Entity'), ('file', 'File'), ('image', 'Image'), ('embed', 'Embed')], default='text', max_length=20, verbose_name='Type'),
        ),
    ]



# -*- coding: utf-8 -*-

__all__ = ["OscarCatalogueAppPluginConfig"]


def __getattr__(name):
    if name != "OscarCatalogueAppPluginConfig":
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
    from .entrypoint import (
        OcyanOscar3CatalogueAppPluginConfig as OscarCatalogueAppPluginConfig,
    )

    return OscarCatalogueAppPluginConfig



# Generated by Django 3.2.13 on 2022-07-08 10:23

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0025_remove_seo_fields'),
        ('catalogue', '0026_predefined_product_options'),
    ]

    operations = [
    ]



from ocyan.core import plugins as core_plugin


class OscarCataloguePluginConfig(core_plugin.OcyanPluginConfig):
    def contribute_to_entity_types(self, entity_types):
        return entity_types


def get_catalogue_plugins():
    return core_plugin.get_plugins(cls=OscarCataloguePluginConfig)


def get_entity_types():
    payment_plugins = get_catalogue_plugins()
    return payment_plugins.pipe("contribute_to_entity_types", [])



"""
Solves some problems that can appear when transactions are not used and
a piece of code breaks, leaving the tree in an inconsistent state.

The problems this method solves are:

   1. Nodes with an incorrect ``depth`` or ``numchild`` values due to
      incorrect code and lack of database transactions.
   2. "Holes" in the tree. This is normal if you move/delete nodes a
      lot. Holes in a tree don't affect performance,
   3. Incorrect ordering of nodes when ``node_order_by`` is enabled.
      Ordering is enforced on *node insertion*, so if an attribute in
      ``node_order_by`` is modified after the node is inserted, the
      tree ordering will be inconsistent.
   4. Fix up parents_are_public state for categories.

:param destructive:

    A boolean value. If True, a more agressive fix_tree method will be
    attempted. If False (the default), it will use a safe (and fast!)
    fix approach, but it will only solve the ``depth`` and
    ``numchild`` nodes, it won't fix the tree holes or broken path
    ordering.

    .. warning::

       Currently what the ``destructive`` method does is:

       1. Backup the tree with :meth:`dump_data`
       2. Remove all nodes in the tree.
       3. Restore the tree with :meth:`load_data`

       So, even when the primary keys of your nodes will be preserved,
       this method isn't foreign-key friendly. That needs complex
       in-place tree reordering, not available at the moment (hint:
       patches are welcome).
"""

from django.core.management.base import BaseCommand

from oscar.core.loading import get_model

Category = get_model("catalogue", "Category")


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument(
            "--destructive",
            action="store_true",
            help="make sure to read the help before you use this",
        )

    def handle(self, *args, **options):
        Category.fix_tree(options["destructive"])



"""
This file is only used in oxyan with oscar 3
"""

# pylint: disable=no-member
from oscar.apps.catalogue import apps

from .shared.appconfig import CatalogueAppConfigMixin


class CatalogueConfig(CatalogueAppConfigMixin, apps.CatalogueConfig):
    name = "ocyan.plugin.oscar_catalogue.catalogue"

    def get_urls(self):
        review_urlpatterns = apps.CatalogueReviewsOnlyConfig.get_urls(self)
        return self.post_process_urls(
            review_urlpatterns + self.new_urls() + self.old_urls()
        )



# Generated by Django 5.2.3 on 2025-07-17 07:01

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0032_category_exclude_from_menu_category_long_description'),
        ('catalogue', '0032_merge_20250424_0937'),
    ]

    operations = [
    ]



from django.core.exceptions import ValidationError
from django.test import TestCase

from oscar.core.loading import get_model

Product = get_model("catalogue", "Product")


class ProductAttributeQuerysetTest(TestCase):
    fixtures = ["catalogue/catalogue"]

    def test_query_multiple_producttypes(self):
        "We should be able to query over multiple product classes"
        result = Product.objects.filter_by_attributes(henkie="bah bah")
        self.assertEqual(result.count(), 2)
        result1, result2 = list(result)

        self.assertNotEqual(result1.product_class, result2.product_class)
        self.assertEqual(result1.attr.henkie, result2.attr.henkie)

    def test_further_filtering(self):
        "The returned queryset should be ready for further filtering"
        result = Product.objects.filter_by_attributes(henkie="bah bah")
        photo = result.filter(title__contains="Photo")
        self.assertEqual(photo.count(), 1)

    def test_empty_results(self):
        "Empty results are possible without errors"
        result = Product.objects.filter_by_attributes(doesnotexist=True)
        self.assertFalse(
            result.exists(), "querying with bulshit attributes should give no results"
        )
        result = Product.objects.filter_by_attributes(henkie="zulthoofd")
        self.assertFalse(
            result.exists(), "querying with non existing values should give no results"
        )
        result = Product.objects.filter_by_attributes(henkie=True)
        self.assertFalse(
            result.exists(), "querying with wring value type should give no results"
        )

    def test_text_value(self):
        result = Product.objects.filter_by_attributes(subtitle="superhenk")
        self.assertTrue(result.exists())
        result = Product.objects.filter_by_attributes(subtitle="kekjo")
        self.assertTrue(result.exists())
        result = Product.objects.filter_by_attributes(subtitle=True)
        self.assertFalse(result.exists())

    def test_formatted_text(self):
        html = '<p style="margin: 0px; font-stretch: normal; font-size: 12px; line-height: normal; font-family: Helvetica;">Vivamus auctor leo vel dui. Aliquam erat volutpat. Phasellus nibh. Vestibulum ante ipsum primis in faucibus orci luctus et ultrices posuere cubilia Curae; Cras tempor. Morbi egestas, <em>urna</em> non consequat tempus, <strong>nunc</strong> arcu mollis enim, eu aliquam erat nulla non nibh. Duis consectetuer malesuada velit. Nam ante nulla, interdum vel, tristique ac, condimentum non, tellus. Proin ornare feugiat nisl. Suspendisse dolor nisl, ultrices at, eleifend vel, consequat at, dolor.</p>'
        result = Product.objects.filter_by_attributes(additional_info=html)
        self.assertTrue(result.exists())

    def test_boolean(self):
        result = Product.objects.filter_by_attributes(available=True)
        self.assertTrue(result.exists())
        result = Product.objects.filter_by_attributes(available=0)
        self.assertTrue(result.exists())
        with self.assertRaises(ValidationError):
            result = Product.objects.filter_by_attributes(available="henk")

    def test_number(self):
        result = Product.objects.filter_by_attributes(facets=4)
        self.assertTrue(result.exists())
        with self.assertRaises(ValueError):
            result = Product.objects.filter_by_attributes(facets="four")

        result = Product.objects.filter_by_attributes(facets=1)
        self.assertFalse(result.exists())

    def test_float(self):
        result = Product.objects.filter_by_attributes(hypothenusa=1.25)
        self.assertTrue(result.exists())
        with self.assertRaises(ValueError):
            result = Product.objects.filter_by_attributes(facets="four")

        result = Product.objects.filter_by_attributes(hypothenusa=1)
        self.assertFalse(result.exists())

    def test_option(self):
        result = Product.objects.filter_by_attributes(kind="totalitarian")
        self.assertTrue(result.exists())
        result = Product.objects.filter_by_attributes(kind=True)
        self.assertFalse(result.exists())

        result = Product.objects.filter_by_attributes(kind="omnimous")
        self.assertFalse(result.exists())

    def test_multi_option(self):
        result = Product.objects.filter_by_attributes(subkinds="megalomane")
        self.assertTrue(result.exists())
        self.assertEqual(result.count(), 2)
        result = Product.objects.filter_by_attributes(subkinds=True)
        self.assertFalse(result.exists())

        result = Product.objects.filter_by_attributes(subkinds="omnimous")
        self.assertFalse(result.exists())

        result = Product.objects.filter_by_attributes(subkinds__contains="om")
        self.assertTrue(result.exists(), "megalomane conains om")
        self.assertEqual(result.count(), 2)

    def test_multiple_attributes(self):
        result = Product.objects.filter_by_attributes(
            subkinds="megalomane", available=True
        )
        self.assertTrue(result.exists())

        result = Product.objects.filter_by_attributes(
            kind="totalitarian",
            hypothenusa=1.25,
            facets=8,
            subtitle="superhenk",
            subkinds="megalomane",
            available=False,
        )
        self.assertTrue(result.exists())

    def test_lookups(self):
        result = Product.objects.filter_by_attributes(facets__lte=4)
        self.assertEqual(result.count(), 1)

        result = Product.objects.filter_by_attributes(facets__lte=8)
        self.assertEqual(result.count(), 2)

        result = Product.objects.filter_by_attributes(facets__lt=8)
        self.assertEqual(result.count(), 1)



# Generated by Django 4.2.11 on 2024-09-11 10:51

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("catalogue", "0028_merge_in_unique_codes"),
        ("catalogue", "0028_product_priority"),
    ]

    operations = []



# Generated by Django 2.0.13 on 2019-10-03 09:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0019_product_priority'),
        ('catalogue', '0018_category_parents_are_public'),
    ]

    operations = [
    ]



# Generated by Django 4.2.15 on 2024-09-11 10:37

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("catalogue", "0027_attributeoption_code_attributeoptiongroup_code_and_more"),
    ]

    """
        We empty this operation because 0019_product_priority already did this,
        but we want this 0028 migration for compatibility with migrating older oscar projects to ocyan
    """
    operations = []



import logging
from collections import defaultdict

from django.db import connection, models
from django.db.models import Exists, OuterRef
from django.db.models.constants import LOOKUP_SEP

from oscar.apps.catalogue import managers as oscar
from oscar.core.loading import get_model

logger = logging.getLogger(__name__)


class AttributeFilter(dict):
    def __init__(self, filter_kwargs):
        super(AttributeFilter, self).__init__()

        num_fields = 0
        for key, value in filter_kwargs.items():
            num_fields += 1
            if LOOKUP_SEP in key:
                field_name, lookup = key.split(LOOKUP_SEP, 1)
                self[field_name] = (lookup, value)
            else:
                self[key] = (None, value)

        self.num_fields = num_fields

    def field_names(self):
        return self.keys()

    def _selector(self, attribute_type):
        if attribute_type == "option" or attribute_type == "multi_option":
            return "attribute_values__value_%s__option" % attribute_type
        else:
            return "attribute_values__value_%s" % attribute_type

    def _select_value(self, types, lookup, value):
        _filter = models.Q()
        for _type in types:
            sel = self._selector(_type)
            if lookup is not None:
                sel = "%s%s%s" % (sel, LOOKUP_SEP, lookup)

            kwargs = dict()
            kwargs[sel] = value
            _filter |= models.Q(**kwargs)

        return _filter

    def fast_query(self, attribute_types, queryset):
        qs = queryset
        typedict = defaultdict(list)

        for code, attribute_type in attribute_types:
            typedict[code].append(attribute_type)

        for code, (lookup, value) in self.items():
            selected_values = self._select_value(typedict[code], lookup, value)
            if (
                not selected_values
            ):  # if no value clause can be formed, no result can be formed.
                return queryset.none()

            qs = qs.filter(selected_values, attribute_values__attribute__code=code)

        return qs


class ProductQuerySet(oscar.ProductQuerySet):
    def count(self):
        if not connection.features.allows_group_by_selected_pks:
            # the count will be super slow if all values are grouped by
            minimum_queryset = self.values("id")
            return super(ProductQuerySet, minimum_queryset).count()

        return super(ProductQuerySet, self).count()

    def base_queryset(self):
        """
        Applies select_related and prefetch_related for commonly related
        models to save on queries
        """
        Option = self.model.product_options.rel.model
        product_class_options = Option.objects.filter(
            productclass=models.OuterRef("product_class")
        )
        product_options = Option.objects.filter(product=models.OuterRef("pk"))
        return (
            self.select_related("product_class")
            .prefetch_related(
                "children",
                "product_options",
                "product_class__options",
                "stockrecords",
                "images",
            )
            .annotate(
                has_product_class_options=models.Exists(product_class_options),
                has_product_options=models.Exists(product_options),
            )
        )

    def filter_by_attributes(self, **filter_kwargs):
        attribute_filter = AttributeFilter(filter_kwargs)

        ProductAttribute = self.model.attributes.rel.model
        attribute_types = ProductAttribute.objects.values_list("code", "type").filter(
            code__in=attribute_filter.field_names()
        )

        return attribute_filter.fast_query(attribute_types, self)

    def browsable(self):
        """
        Excludes products without categories. They are used in product bundles.
        """
        ProductCategory = get_model("catalogue", "ProductCategory")
        qs = super().browsable()
        # Products without categories are used for product options only.
        return qs.filter(Exists(ProductCategory.objects.filter(product=OuterRef("pk"))))


ProductManager = ProductQuerySet.as_manager



from django.urls import reverse
from django.views.generic import RedirectView


class CategoryRedirectView(RedirectView):
    permanent = True
    query_string = True

    def get_redirect_url(self, *args, **kwargs):
        # Use reverse to build the URL, which handles encoding properly
        return reverse("catalogue:category", kwargs=kwargs)


class CatalogueIndexRedirectView(RedirectView):
    permanent = True
    query_string = True

    def get_redirect_url(self, *args, **kwargs):
        # Use reverse to build the URL for the catalogue index
        return reverse("catalogue:index")



from django.db import models
from django.utils.translation import gettext_lazy as _


class AbstractSEOFields(models.Model):
    seo_title = models.CharField(
        _("SEO Meta title"),
        max_length=255,
        help_text=_(
            "SEO Meta titles should be less than 60 characters, if left empty the product title will be used"
        ),
        blank=True,
        null=True,
    )
    seo_description = models.CharField(
        _("SEO Meta description"),
        max_length=1024,
        help_text=_(
            "SEO Meta descriptions should be less than 300 characters, if left empty the product description will be used"
        ),
        blank=True,
        null=True,
    )
    seo_keywords = models.CharField(
        _("SEO Meta keywords"),
        max_length=1024,
        help_text=_(
            "Meta Keywords don't make any improvements in SEO, use Meta description instead."
        ),
        blank=True,
        null=True,
    )

    class Meta:
        abstract = True



# Generated by Django 2.0.13 on 2019-07-18 10:25

from django.db import migrations, models
import django.core.validators

class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0018_ocyan_additions'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='product',
            options={'ordering': ['-priority', '-date_created'], 'verbose_name': 'Product', 'verbose_name_plural': 'Products'},
        ),
        migrations.AddField(
            model_name='product',
            name='priority',
            field=models.SmallIntegerField(db_index=True, default=0, help_text='The highest priority products are shown first', verbose_name='Priority', validators=[django.core.validators.MinValueValidator(-1)]),
        ),
    ]



from ocyan.core.fender import config

DEFAULT_SLUG = "product"
MINIMIM_QUANTITY_ATTRIBUTE = config.get(
    "oscar_catalogue", "minimum_quantity_attribute_code", "min_quantity"
)
SLUG_ID_SEPARATOR = config.get("oscar_catalogue", "slug_id_separator", "-")



import os
from decimal import Decimal

import lxml.html
from django.db import models
from django.db.models.functions import Coalesce
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from oscar.apps.catalogue import abstract_models as oscar
from oscar.core.loading import get_class
from oscar.models.fields import AutoSlugField

from ocyan.plugin.oscar.fender import ALL_TAX_RATES
from ocyan.plugin.oscar.fender import get_default_tax_rate as vat_default

from ocyan.plugin.oscar_catalogue.constants import (
    DEFAULT_SLUG,
    MINIMIM_QUANTITY_ATTRIBUTE,
)

tax_rate_as_fraction = get_class("taxes.taxrate", "tax_rate_as_fraction")
AbstractSEOFields = get_class("catalogue.helpers", "AbstractSEOFields")


class AbstractProductClass(oscar.AbstractProductClass):
    slug = AutoSlugField(
        _("Slug"), max_length=128, unique=True, populate_from="name", editable=True
    )

    class Meta(oscar.AbstractProductClass.Meta):
        abstract = True


class AbstractOption(oscar.AbstractOption):
    code = AutoSlugField(
        _("Code"), max_length=128, unique=True, populate_from="name", editable=True
    )
    order = models.IntegerField(
        _("Ordering"),
        null=True,
        blank=True,
        help_text=_("Controls the ordering of product options on product detail pages"),
        db_index=True,
    )

    class Meta(oscar.AbstractOption.Meta):
        abstract = True
        ordering = ["order", "name"]


class AbstractCategory(oscar.AbstractCategory, AbstractSEOFields):
    class Meta(oscar.AbstractCategory.Meta):
        abstract = True


class AbstractProductAttribute(oscar.AbstractProductAttribute):
    EMBED = "embed"

    TYPE_CHOICES = oscar.AbstractProductAttribute.TYPE_CHOICES + ((EMBED, _("Embed")),)

    type = models.CharField(
        choices=TYPE_CHOICES,
        default=TYPE_CHOICES[0][0],
        max_length=20,
        verbose_name=_("Type"),
    )

    class Meta(oscar.AbstractProductAttribute.Meta):
        abstract = True

    # pylint: disable=protected-access
    _validate_embed = oscar.AbstractProductAttribute._validate_text


class AbstractProductAttributeValue(oscar.AbstractProductAttributeValue):
    value_embed = models.TextField(
        _("Youtube Video ID"),
        blank=True,
        null=True,
        help_text=_(
            "This can be found in the URL of a video. So for example: https://www.youtube.com/watch?v=dkIU-ygvtps. In this case {dkIU-ygvtps} is the Youtube Video ID"
        ),
    )

    class Meta(oscar.AbstractProductAttributeValue.Meta):
        abstract = True

    @property
    def _embed_as_html(self):
        return render_to_string(
            "oscar/catalogue/partials/embed.html", {"value_embed": self.value}
        )

    @property
    def _file_as_html(self):
        return render_to_string(
            "oscar/catalogue/partials/file.html",
            {"value_file": self.value, "display_name": self._file_as_text},
        )

    @property
    def _image_as_html(self):
        return render_to_string(
            "oscar/catalogue/partials/image.html",
            {"value_image": self.value, "display_name": self._image_as_text},
        )

    @property
    def _file_as_text(self):
        return os.path.basename(self.value.name)

    @property
    def _image_as_text(self):
        return os.path.basename(self.value.name)


class AbstractProduct(oscar.AbstractProduct, AbstractSEOFields):
    vat_rate = models.CharField(
        _("VAT rate"),
        max_length=5,
        choices=ALL_TAX_RATES,
        default=vat_default(),
        null=True,
    )

    long_description = models.TextField(
        _("Content"),
        blank=True,
        null=True,
        help_text=_("A far more elaborate, but optional description of the product"),
    )

    @property
    def min_quantity(self):
        min_quantity_values = self.get_attribute_values().filter(
            attribute__code=MINIMIM_QUANTITY_ATTRIBUTE
        )
        if min_quantity_values.exists():
            return min_quantity_values.first().value

        return 1

    @property
    def tax_percentage(self):
        return tax_rate_as_fraction(self.vat_rate)

    @property
    def variants_ordered_by_price(self):
        return (
            self.children.filter(is_public=True)
            .annotate(
                lowest_price=Coalesce(
                    models.Avg("stockrecords__price_excl_tax"),
                    models.Value(
                        Decimal(999999999)
                    ),  # put children with no stock at the end
                )
            )
            .order_by("lowest_price", "-priority")
        )

    @property
    def variants_ordered_for_admin(self):
        return (
            self.children.all()
            .annotate(
                lowest_price=Coalesce(
                    models.Avg("stockrecords__price_excl_tax"),
                    models.Value(
                        Decimal(999999999)
                    ),  # put children with no stock at the end
                )
            )
            .order_by("lowest_price", "-priority")
        )

    @property
    def options(self):
        """
        TODO: Is this really needed????
        Override to give precedence to options defined on product over product_class
        """
        if self.product_options.exists():
            return self.product_options.distinct()

        return self.get_product_class().options.distinct()

    @cached_property
    def has_options(self):
        # Extracting annotated value with number of product class options
        # from product list queryset.
        has_product_class_options = getattr(self, "has_product_class_options", None)
        has_product_options = getattr(self, "has_product_options", None)
        if has_product_class_options is not None and has_product_options is not None:
            return has_product_class_options or has_product_options
        # this can be changed to self.options.exists() if we revert to the default implementation.
        return (
            self.get_product_class().options.all() | self.product_options.all()
        ).exists()

    @property
    def sorted_recommended_products(self):
        """Keeping order by recommendation ranking."""
        return [
            r.recommendation
            for r in self.primary_recommendations.filter(recommendation__is_public=True)
            .select_related("recommendation")
            .all()
        ]

    def primary_category(self):
        return self.categories.first()

    def get_absolute_url(self):
        slug = self.slug or DEFAULT_SLUG
        return reverse("catalogue:detail", kwargs={"product_slug": slug, "pk": self.pk})

    class Meta(oscar.AbstractProduct.Meta):
        abstract = True
        ordering = ["-priority", "-date_created"]

    def clean(self):
        super().clean()

        # We'll clean out html fields, so they don't have an empty p or whatever.
        # This allows us to use just check for product.description et al.
        for name in ["description", "long_description"]:
            value = getattr(self, name) or ""
            value = value.strip()  # pylint: disable=no-member
            if "<img" in value:
                continue
            if value:
                tree = lxml.html.fromstring(value)
                if not tree.text_content().strip():
                    value = ""
            setattr(self, name, value)


class AbstractEntityAttributeValue(models.Model):
    """
    This class can be used to implement entities that can be picked in
    the dashboard and still work in oscarapi and the import export plugin.
    """

    name = models.CharField(_("Name"), max_length=125)
    code = models.SlugField(_("Code"), max_length=140, unique=True)

    def __str__(self):
        return self.name

    def json(self):
        return self.code

    class Meta:
        abstract = True



# -*- coding: utf-8 -*- pylint: disable=E1120
from decimal import Decimal as D
from unittest import mock

from django.contrib.auth import get_user_model
from django.test import RequestFactory, TestCase, override_settings
from django.urls import reverse

from oscar.core.loading import get_model

from oscarapi.basket.operations import get_user_basket, prepare_basket

import ocyan.core.fender
import ocyan.plugin.oscar.taxes.taxrate
from ocyan.plugin.oscar.taxes.utils import (
    exclusive_tax_component,
    tax_from_price_incl_tax,
)

Basket = get_model("basket", "Basket")
Product = get_model("catalogue", "Product")
User = get_user_model()
factory = RequestFactory()

OCYAN_MOCK_CONFIG_INCL_TAX = {"oscar": {"use_price_incl_tax": True}}
OCYAN_MOCK_CONFIG_EXCL_TAX = {"oscar": {"use_price_incl_tax": False}}


# Mocking method
def get_tax_rate_as_fraction_mocked_defaults():
    """
    All tests below are tested with the NL vat country in mind. Projects having a different value
    set will have failing tests, before this mock we would have added those tests in the skipped_tests list.
    Instead of skipping the tests, it's better to just mock the value to NL, as there are tests in other plugins
    that test with different OSCAR_VAT_COUNTRY_OF_ORIGIN_CODE.

    Django's override_settings does not work when a setting is used as default method argument, eg;
    def foo(default=settings.SOME_SETIING)

    So that's why we're mocking this default argument with mock.patch.
    """
    defaults = list(ocyan.plugin.oscar.taxes.taxrate.tax_rate_as_fraction.__defaults__)
    defaults[0] = "NL"
    return tuple(defaults)


tax_rate_as_fraction_mocked_defaults = get_tax_rate_as_fraction_mocked_defaults()


@override_settings(OSCAR_VAT_COUNTRY_OF_ORIGIN_CODE="NL")
@mock.patch.object(
    ocyan.plugin.oscar.taxes.taxrate.tax_rate_as_fraction,
    "__defaults__",
    tax_rate_as_fraction_mocked_defaults,
)
class TaxTest(TestCase):
    fixtures = [
        "catalogue/catalogue",
        "catalogue/auth",
    ]

    def assertLessThanOneCentDifference(self, calculated, basket):
        self.assertLessEqual(abs(calculated - basket), D("0.01"))

    def test_catalogue_initialized(self):
        "The catalogue should be initialized"
        self.assertEqual(Product.objects.count(), 4)
        self.assertEqual(User.objects.count(), 2)

    @mock.patch.object(ocyan.core.fender.config, "settings", OCYAN_MOCK_CONFIG_INCL_TAX)
    def test_add_product_to_basket_incl_tax(self):
        admin = User.objects.get(username="admin")
        self.client.force_login(admin)
        basket = get_user_basket(admin)
        self.assertEqual(Basket.objects.count(), 1)

        self.assertEqual(basket.total_excl_tax_excl_discounts, D("0.00"))
        self.assertEqual(basket.total_incl_tax_excl_discounts, D("0.00"))
        self.assertEqual(basket.total_discount, D("0.00"))
        self.assertEqual(basket.total_excl_tax, D("0.00"))
        self.assertEqual(basket.total_tax, D("0.00"))
        self.assertEqual(basket.total_incl_tax, D("0.00"))

        basket_add_url = reverse("basket:add", args=(2,))
        body = {"quantity": 1}

        response = self.client.post(basket_add_url, body)

        request = factory.post(basket_add_url, body)
        request.user = admin
        request.session = self.client.session

        basket.refresh_from_db()
        basket = prepare_basket(basket, request)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(Basket.objects.count(), 1)

        self.assertEqual(basket.total_excl_tax_excl_discounts, D("330.58"))
        self.assertEqual(basket.total_incl_tax_excl_discounts, D("400.00"))
        self.assertEqual(basket.total_discount, D("0.00"))
        self.assertEqual(basket.total_excl_tax, D("330.58"))
        self.assertEqual(basket.total_tax, D("69.42"))
        self.assertEqual(basket.total_incl_tax, D("400.00"))
        self.assertEqual(
            tax_from_price_incl_tax(basket.total_incl_tax, D(".21")), D("69.4215")
        )
        self.assertAlmostEqual(
            exclusive_tax_component(D("0.21")) * basket.total_excl_tax,
            basket.total_incl_tax,
            2,
        )

    @mock.patch.object(ocyan.core.fender.config, "settings", OCYAN_MOCK_CONFIG_INCL_TAX)
    def test_add_multiple_products_to_the_freaking_basket_incl_tax(self):
        "Adding mutltiple products should still calculate tax correctly"
        self.test_add_product_to_basket_incl_tax()

        admin = User.objects.get(username="admin")
        basket = get_user_basket(admin)

        basket_add_url = reverse("basket:add", args=(3,))
        body = {"quantity": 2}

        response = self.client.post(basket_add_url, body)

        request = factory.post(basket_add_url, body)
        request.user = admin
        request.session = self.client.session

        basket.refresh_from_db()
        basket = prepare_basket(basket, request)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(Basket.objects.count(), 1)

        self.assertEqual(basket.total_excl_tax_excl_discounts, D("368.60"))
        self.assertEqual(basket.total_incl_tax_excl_discounts, D("446.00"))
        self.assertEqual(basket.total_discount, D("0.00"))
        self.assertEqual(basket.total_excl_tax, D("368.60"))
        self.assertEqual(basket.total_tax, D("77.40"))
        self.assertEqual(basket.total_incl_tax, D("446.00"))
        self.assertLessThanOneCentDifference(
            exclusive_tax_component(D("0.21")) * basket.total_excl_tax,
            basket.total_incl_tax,
        )

    @mock.patch.object(ocyan.core.fender.config, "settings", OCYAN_MOCK_CONFIG_EXCL_TAX)
    def test_add_product_to_basket_excl_tax(self):
        admin = User.objects.get(username="admin")
        self.client.force_login(admin)
        basket = get_user_basket(admin)
        self.assertEqual(Basket.objects.count(), 1)

        self.assertEqual(basket.total_excl_tax_excl_discounts, D("0.00"))
        self.assertEqual(basket.total_incl_tax_excl_discounts, D("0.00"))
        self.assertEqual(basket.total_discount, D("0.00"))
        self.assertEqual(basket.total_excl_tax, D("0.00"))
        self.assertEqual(basket.total_tax, D("0.00"))
        self.assertEqual(basket.total_incl_tax, D("0.00"))

        basket_add_url = reverse("basket:add", args=(2,))
        body = {"quantity": 1}

        response = self.client.post(basket_add_url, body)

        request = factory.post(basket_add_url, body)
        request.user = admin
        request.session = self.client.session

        basket.refresh_from_db()
        basket = prepare_basket(basket, request)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(Basket.objects.count(), 1)

        self.assertEqual(basket.total_excl_tax_excl_discounts, D("400.00"))
        self.assertEqual(basket.total_incl_tax_excl_discounts, D("484.00"))
        self.assertEqual(basket.total_discount, D("0.00"))
        self.assertEqual(basket.total_excl_tax, D("400.00"))
        self.assertEqual(basket.total_tax, D("84.00"))
        self.assertEqual(basket.total_incl_tax, D("484.00"))
        self.assertEqual(
            tax_from_price_incl_tax(basket.total_incl_tax, D(".21")), basket.total_tax
        )
        self.assertAlmostEqual(
            exclusive_tax_component(D("0.21")) * basket.total_excl_tax,
            basket.total_incl_tax,
            2,
        )

    @mock.patch.object(ocyan.core.fender.config, "settings", OCYAN_MOCK_CONFIG_EXCL_TAX)
    def test_add_multiple_products_to_the_freaking_basket_excl_tax(self):
        self.test_add_product_to_basket_excl_tax()

        admin = User.objects.get(username="admin")
        basket = get_user_basket(admin)

        basket_add_url = reverse("basket:add", args=(3,))
        body = {"quantity": 2}

        response = self.client.post(basket_add_url, body)

        request = factory.post(basket_add_url, body)
        request.user = admin
        request.session = self.client.session

        basket.refresh_from_db()
        basket = prepare_basket(basket, request)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(Basket.objects.count(), 1)

        self.assertEqual(basket.total_excl_tax_excl_discounts, D("446.00"))
        self.assertEqual(basket.total_incl_tax_excl_discounts, D("539.66"))
        self.assertEqual(basket.total_discount, D("0.00"))
        self.assertEqual(basket.total_excl_tax, D("446.00"))
        self.assertEqual(basket.total_tax, D("93.66"))
        self.assertEqual(basket.total_incl_tax, D("539.66"))
        self.assertLessThanOneCentDifference(
            exclusive_tax_component(D("0.21")) * basket.total_excl_tax,
            basket.total_incl_tax,
        )


@override_settings(OSCAR_VAT_COUNTRY_OF_ORIGIN_CODE="NL")
@mock.patch.object(
    ocyan.plugin.oscar.taxes.taxrate.tax_rate_as_fraction,
    "__defaults__",
    tax_rate_as_fraction_mocked_defaults,
)
class OfferTest(TaxTest):
    fixtures = TaxTest.fixtures + ["catalogue/offer"]

    @mock.patch.object(ocyan.core.fender.config, "settings", OCYAN_MOCK_CONFIG_INCL_TAX)
    def test_add_multiple_products_to_the_freaking_basket_incl_tax(self):
        with self.settings(OSCAR_OFFERS_INCL_TAX=True):
            self.test_add_product_to_basket_incl_tax()

            admin = User.objects.get(username="admin")
            basket = get_user_basket(admin)

            basket_add_url = reverse("basket:add", args=(3,))
            body = {"quantity": 2}

            response = self.client.post(basket_add_url, body)

            request = factory.post(basket_add_url, body)
            request.user = admin
            request.session = self.client.session

            basket.refresh_from_db()
            basket = prepare_basket(basket, request)

            self.assertEqual(response.status_code, 302)
            self.assertEqual(Basket.objects.count(), 1)

            self.assertEqual(basket.total_excl_tax_excl_discounts, D("368.60"))
            self.assertEqual(basket.total_incl_tax_excl_discounts, D("446.00"))
            self.assertEqual(basket.total_discount, D("4.60"))
            self.assertEqual(basket.total_excl_tax, D("364.79"))
            self.assertEqual(basket.total_tax, D("76.61"))
            self.assertEqual(basket.total_incl_tax, D("441.40"))
            self.assertLessThanOneCentDifference(
                exclusive_tax_component(D("0.21")) * basket.total_excl_tax,
                basket.total_incl_tax,
            )

    @mock.patch.object(ocyan.core.fender.config, "settings", OCYAN_MOCK_CONFIG_EXCL_TAX)
    def test_add_multiple_products_to_the_freaking_basket_excl_tax(self):
        with self.settings(OSCAR_OFFERS_INCL_TAX=False):
            self.test_add_product_to_basket_excl_tax()

            admin = User.objects.get(username="admin")
            basket = get_user_basket(admin)

            basket_add_url = reverse("basket:add", args=(3,))
            body = {"quantity": 2}

            response = self.client.post(basket_add_url, body)

            request = factory.post(basket_add_url, body)
            request.user = admin
            request.session = self.client.session

            basket.refresh_from_db()
            basket = prepare_basket(basket, request)

            self.assertEqual(response.status_code, 302)
            self.assertEqual(Basket.objects.count(), 1)

            self.assertEqual(basket.total_excl_tax_excl_discounts, D("446.00"))
            self.assertEqual(basket.total_incl_tax_excl_discounts, D("539.66"))
            self.assertEqual(basket.total_discount, D("4.60"))
            self.assertEqual(basket.total_excl_tax, D("441.40"))
            self.assertEqual(basket.total_tax, D("92.69"))
            self.assertEqual(basket.total_incl_tax, D("534.09"))

            self.assertAlmostEqual(
                exclusive_tax_component(D("0.21")) * basket.total_excl_tax,
                basket.total_incl_tax,
                2,
            )

    @mock.patch.object(ocyan.core.fender.config, "settings", OCYAN_MOCK_CONFIG_EXCL_TAX)
    def test_add_product_to_basket_with_flat_discount_excl_tax(self):
        with self.settings(OSCAR_OFFERS_INCL_TAX=False):
            admin = User.objects.get(username="admin")
            self.client.force_login(admin)
            basket = get_user_basket(admin)
            self.assertEqual(Basket.objects.count(), 1)

            self.assertEqual(basket.total_excl_tax_excl_discounts, D("0.00"))
            self.assertEqual(basket.total_incl_tax_excl_discounts, D("0.00"))
            self.assertEqual(basket.total_discount, D("0.00"))
            self.assertEqual(basket.total_excl_tax, D("0.00"))
            self.assertEqual(basket.total_tax, D("0.00"))
            self.assertEqual(basket.total_incl_tax, D("0.00"))

            basket_add_url = reverse("basket:add", args=(4,))
            body = {"quantity": 2}

            response = self.client.post(basket_add_url, body)

            request = factory.post(basket_add_url, body)
            request.user = admin
            request.session = self.client.session

            basket.refresh_from_db()
            basket = prepare_basket(basket, request)

            self.assertEqual(response.status_code, 302)
            self.assertEqual(Basket.objects.count(), 1)

            self.assertEqual(basket.total_excl_tax_excl_discounts, D("468.00"))
            self.assertEqual(basket.total_incl_tax_excl_discounts, D("566.28"))

            self.assertEqual(basket.total_discount, D("10.00"))
            self.assertEqual(basket.total_tax, D("96.18"))
            self.assertEqual(basket.total_incl_tax, D("554.18"))
            self.assertEqual(basket.total_excl_tax, D("458.00"))
            self.assertEqual(
                tax_from_price_incl_tax(basket.total_incl_tax, D(".21")),
                basket.total_tax,
            )
            self.assertLessThanOneCentDifference(
                exclusive_tax_component(D("0.21")) * basket.total_excl_tax,
                basket.total_incl_tax,
            )

    @mock.patch.object(ocyan.core.fender.config, "settings", OCYAN_MOCK_CONFIG_INCL_TAX)
    def test_add_product_to_basket_with_flat_discount_incl_tax(self):
        with self.settings(OSCAR_OFFERS_INCL_TAX=True):
            admin = User.objects.get(username="admin")
            self.client.force_login(admin)
            basket = get_user_basket(admin)
            self.assertEqual(Basket.objects.count(), 1)

            self.assertEqual(basket.total_excl_tax_excl_discounts, D("0.00"))
            self.assertEqual(basket.total_incl_tax_excl_discounts, D("0.00"))
            self.assertEqual(basket.total_discount, D("0.00"))
            self.assertEqual(basket.total_excl_tax, D("0.00"))
            self.assertEqual(basket.total_tax, D("0.00"))
            self.assertEqual(basket.total_incl_tax, D("0.00"))

            basket_add_url = reverse("basket:add", args=(4,))
            body = {"quantity": 2}

            response = self.client.post(basket_add_url, body)

            request = factory.post(basket_add_url, body)
            request.user = admin
            request.session = self.client.session

            basket.refresh_from_db()
            basket = prepare_basket(basket, request)

            self.assertEqual(response.status_code, 302)
            self.assertEqual(Basket.objects.count(), 1)

            self.assertEqual(basket.total_excl_tax_excl_discounts, D("386.78"))
            self.assertEqual(basket.total_incl_tax_excl_discounts, D("468.00"))
            self.assertEqual(basket.total_discount, D("10.00"))
            self.assertEqual(basket.total_incl_tax, D("458.00"))
            self.assertEqual(basket.total_excl_tax, D("378.51"))
            self.assertEqual(basket.total_tax, D("79.49"))
            self.assertLessThanOneCentDifference(
                tax_from_price_incl_tax(basket.total_incl_tax, D(".21")),
                basket.total_tax,
            )
            self.assertLessThanOneCentDifference(
                exclusive_tax_component(D("0.21")) * basket.total_excl_tax,
                basket.total_incl_tax,
            )


class ProductTest(TestCase):
    fixtures = ["catalogue/catalogue"]

    def test_clean_whitespace(self):
        values = ["<p></p>", " ", "<p> </p>", "<div> <p> </p></div>"]
        product = Product.objects.first()
        for value in values:
            product.description = value
            product.long_description = value
            product.clean()
            self.assertEqual("", product.description)
            self.assertEqual("", product.long_description)

    def test_clean_non_whitespace(self):
        values = [
            "<p>a</p>",
            " a ",
            "<p> a </p>",
            "<div> <p> a </p></div>",
            "<div> <img alt='some image' /> </div>",
        ]
        product = Product.objects.first()
        for value in values:
            product.description = value
            product.long_description = value
            product.clean()
            self.assertEqual(value.strip(), product.description)
            self.assertEqual(value.strip(), product.long_description)



# Generated by Django 2.2.13 on 2021-01-22 12:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0021_option_order'),
    ]

    operations = [
        migrations.AlterField(
            model_name='option',
            name='order',
            field=models.IntegerField(blank=True, db_index=True, help_text='Controls the ordering of product options on product detail pages', null=True, verbose_name='Ordering'),
        ),
    ]
